#ifndef BOOK_HH
#define BOOK_HH

/* A memory mapped BabyChess book.  See etc/BOOK. 
 */

#include "game.hh"
#include "inttypes.hh"
#include "../config.hh"

enum {bo_position= (1 << 0),   /* entry types */ 
      bo_move=     (1 << 1),
      bo_tag=      (1 << 2),
      bo_string=   (1 << 3),

      bo_type=     (1 << 4) - 1,

      bo_check=    (1 << 4),   /* POSITION:  whether there is check */ 
};

class Book
{
public:
	typedef int32_t         sig_t;
	typedef uint32_t       usig_t;

#if HAVE_FAST_INTTYPES
	typedef int_fast32_t   sigf_t;
	typedef uint_fast32_t usigf_t;
#else
	typedef sig_t          sigf_t;
	typedef usig_t        usigf_t; 
#endif 

	enum {sig_size= sizeof(sig_t)};

	/* in sync with sizeof (Entry) */ 
	enum {entry_shift= 5,
	      entry_size= 32}; 

	struct Entry
	{
		union
		{
			usig_t flags, magic;
		};
		union
		{
			struct 
			{
				union
				{
					usig_t dummy;
					char version;
				};
				usig_t endianness, size, first, starting, tags, firstn;
			} header;
			struct 
			{
				usig_t last, current, comment; /* pointers */ 
				sig_t score, nag;
				sig_t ep, halfmove_clock;
			} position;
			struct 
			{
				usig_t cnext, cprevious, lnext,
					lprevious;/*moves */
				usig_t move; /* in "small" format */ 
				usig_t before, after; /* positions */ 
			} move;
			struct
			{
				usig_t next, previous;
				usig_t name, value; /* string */ 
				char name_string[3 * sig_size];
			} tag;
			struct 
			{
				usig_t next, previous;
				char text[5 * sig_size];
			} string;
		};
	};

	Book()
		:  entries(NULL) 
		{
			assert (sizeof (Entry) == entry_size);
			assert (entry_size == (1 << entry_shift)); 
			/* in sync with entry_shift above */
		}
	~Book() {if(entries) Close();}

	bool Is() const {return entries; }
	bool Open_New (const char *filename);
	bool Open_Read(const char *filename); 
	bool Open_Anonymous();
	bool Merge(const Game &game);
	void Merge(const Variation &variation); 
	/* The variation must begin at the current position. 
	 */
	bool Merge_Into_Game(Game &game);
	void Close();
	void Play(const Move &move);
	void Quit_Variation();
	bool Go_Backward();
	/* return whether the move was the main variation */ 
	void Go_To_Start();
	void Go_Forward();
	bool Go_Forward(const Move &move);
	String Comment(sigf_t position= 0);
	/*If 0, current position */
	const Entry &operator [] (unsigned i) const {return entries[i]; }
	const Entry &operator () () const {return entries[current]; }
	void Prepend_Tag(String name, String value);
	void Set_Score(usigf_t position, sigf_t score)
		{
			assert (position < entries[0].header.size);
			assert (entries[position].flags & bo_position);
			entries[position].position.score= score; 
		}
	sigf_t Get_Move(move_t move) const;
	String Get_String(usigf_t i) const;

private:
	Entry *entries;  /* The book is open when not NULL. */
	int fd; 
	/* valid and open when the book is open, or <0 when open as an
	 * anonymous book and mremap() is available.  */  

	usigf_t current; /* current position */ 
	Position position;
	int page_size, page_shift;
	/* set in Open_*() */ 
	int page_entries, page_entries_shift;
	/* Entries per page.  Set in th Open_*() functions.  */
	unsigned pages;
	/* Number of pages used. */
	int prot; /* for mmap() */ 

	void Reset_Position();
	/* set POSITION to the starting position of the book. */ 
	usigf_t Alloc(int count= 1);
	/* return index of first allocated entry */ 
	usigf_t Alloc_String(String text);
	/* with PREVIOUS not set */ 
	void Set_Sizes();
	void Play(usigf_t move);
	void Clear(usigf_t i)
		{ memset(entries+i, '\0', sizeof(Entry));  }
	void Initialize();
	void Merge_Current_Variation_Into_Game(Game &);
};

#endif /* ! BOOK_HH */
