#ifndef DOCUMENT_EDITOR_HH
#define DOCUMENT_EDITOR_HH

#include "file_dialog.hh"
#include "stream_file.hh"
#include "ask_dialog.hh"
#include "menubar.hh"
#include "process_open_info.hh"
#include "toolbar.hh"
#include "window.hh"
#include "editor_window.hh"
#include "packer.hh"
#include "document.hh"

/* This functions is implemented with "Type *" as the second argument
 * in some modules that use this module.  */
Inline bool Check_Filename(String, void *)
{
	return true;
}

class Document_Editor_Base
	:  public Document,
	   public Window,
	   virtual public Editor_Window_Base
{
public:

	typedef Process_Open_Info Open_Info; 

	void DO_Filename_Changed(String filename);
	void DO_Save_Error();
	void DO_Close();

	void MR_Receive(String);

	void Append_Menu(Menubar &menubar);
	void Append_Tool_Buttons(Toolbar &);
	void Set_Valid(bool new_valid, Menubar &menubar)
		{
			valid= new_valid;
			Update_Valid(menubar);
		}

protected:

	virtual bool DE_Set_Args(Open_Info &open_info,
				 bool &anonymous);
		/* Called before opening to set args. OPEN_INFO contains the
		 * program name, and all options that were parsed should be
		 * removed.  Return FALSE if the object should not be opened
		 * (wrong options etc...)  The object also has to remember
		 * somehow that options were passed.  (e.g. EW_Set_Value()
		 * should be ignored later on if the options specified an
		 * initial value (e.g. a position in FEN).  Set ANONYMOUS to
		 * true when encountering the -a option. 
		 */

	void Close_Dialogs();

	Menubar menubar; 
	File_Dialog file_dialog;
	int filename_for;
	/* 0: open
	 * 1: load
	 * 2: save
	 */
	Ask_Dialog ask_dialog; 
	V_Packer packer; 

private:

	GtkWidget *ok_button, *apply_button; 
	bool valid; 

	virtual const char *DE_Mode() const= 0;
	/* Return the mode passed to the BabyChess executable to edit programs
	 * of this type.  
	 */
	virtual const char *DE_Text_New() const= 0;
	virtual const char *DE_Text_Open() const= 0;
	virtual const char *DE_Text_Save() const= 0;
	virtual bool Document_Editor_Open(Pointer <Packer> parent)= 0;
	virtual GdkPoint DE_Window_Size() const;
	virtual void DE_Reset()= 0;
	/* Clear object when window is open. */

	String Wdg_Init_Caption() const;


	void DO_Load_Error(String filename);
	void DO_Ask_Load_Name();
	void DO_Ask_Save_Name(String initial_name);
	void DO_Ask(int);
	bool DO_Use_Filename() const;
	void DO_Save_Close();
	void DO_Reset();

	bool EW_Open();
	bool EW_Transient() const;
	void EW_Apply();
		/* When the value is applied.   */
	bool EW_Reset();

	bool Wdg_Close_Query();

	void Update_Valid(Menubar &);
};

template <class Type>
class Document_Editor
	:  public Document_Editor_Base,
	   public Editor_Window <Type> 
{
public:

	bool Open(Open_Info args)
		/* Open with no owner */
		{
			Verbose_Function();
			Verbose_Int(args.argc);

			bool anonymous= false;

			/* Each editor may have it's own special arguments */ 
			if (! DE_Set_Args(args, anonymous))     return false; 

			Verbose_Int(args.argc);

			/* Strip the program name from ARGS */
			--args.argc;
			++args.argv; 

			/* Without arguments, open a new document */ 
			if (args.argc == 0)          
			{
				return Editor_Window <Type> ::Open("");
			}

			/* Else assume a filename */ 
			if (! Load_File(args.argv[0], anonymous))
				return false;
			
			if (! Editor_Window <Type> ::Open
			    (-1, NULL, this->EW_Get_Value(), NULL, "", true))
				return false;

			if (! anonymous)
				DO_Filename_Changed(Document::Filename()); 
			else 
				Modify();

			return true; 
		}

	bool Open(String msg_edited,
		  Message_Receiver &owner, 
		  Pointer <const Type> initial_value= NULL,
		  Pointer <Widget> parent_window= NULL,
		  String caption= "")
		/* with owner */
		{
			return Editor_Window <Type> ::Open
				(msg_edited, 
				 owner,
				 initial_value,
				 parent_window,
				 caption); 
		}

	void Wdg_Close()
		{
			Close_Dialogs();
			Editor_Window <Type> ::Wdg_Close();
		}

private:
	bool DO_Save(String filename)
		{
			return Write_To_File(filename(),
					     this->EW_Get_Value()); 
		}

	Document::Load_Result DO_Load(String filename)
		{
			Type object;
			Document::Load_Result result;

			/* Check whether filename matches the usual filename for
			 * that data type.  */
			bool cf= Check_Filename(filename, (Type *)NULL);
//			if (Check_Filename(filename, (Type *)NULL))
			Verbose_Int(cf);
			if (cf)
			{
				if (! Read_From_File(filename(), object))
					result= Document::lr_error;
				else
					result= Document::lr_success;
			}

			/* Try to load the file as something else */
			else if (EW_Try_Load(filename, object))
				result= Document::lr_wrong_type;

			/* Try loading anyway */
			else
			{
				if (! Read_From_File(filename(), object))
					result= Document::lr_error;
				else
					result= Document::lr_success;
			}

			if (result != Document::lr_error)
			{
				if (Widget::Is())
				{
					if (! EW_Change_Value(object))
					{
						result= Document::lr_error;
					}
				}
				else
				{
					EW_Set_Value(object); 
				}
			}

			return result;
		}
};

#endif /* ! DOCUMENT_EDITOR_HH */
