#ifndef NOTATION_HH
#define NOTATION_HH

/* Chess notation */

#include "string.hh"
#include "move_name.hh"
#include "piece.hh"

enum /* options */ 
{
	/* type */ 
	nt_short             = (1 << 0),  /* short                                     */
	nt_descriptive       = (1 << 9),  /* descriptive                               */
	nt_nolong            = nt_short | nt_descriptive,
	/* else "long" */ 

	/* bits */ 
	nt_pawn_sign         = (1 << 1),  /* Sign for pawn moves                       */
	nt_ep                = (1 << 2),  /* En passant                                */
	nt_check             = (1 << 3),  /* Check, checkmate                          */
	nt_short_capture     = (1 << 4),  /* Capture in short notation                 */ 
	nt_short_capture_end = (1 << 5),  /* Capture sign at the end in short notation */
	nt_short_pawn        = (1 << 6),  /* Sign for pawn captures in short notation  */ 
	nt_capture_name      = (1 << 7),  /* Name of captured piece in long notation   */ 
	nt_double_check      = (1 << 8),  /* Show double check using "++"              */

	nt_base=  1,
	nt_count= 8,
	nt_used= 10,
};

enum /* text options */ 
{
	/*                             Description        PGN equiv.*/
	s_ep,                       /* en passant         e.p.      */
	s_castling,   
	s_short_castling=s_castling,/* short castling     O-O       */
	s_long_castling,            /* long  castling     O-O-O     */
	s_P,                        /* pieces             P, N, etc.*/
	s_capture= s_P+6,           /* capture sign       x         */
	s_promotion,                /* promotion sign     =         */
	s_color,                    /* white piece        w         */
	s_color_b,                  /* black piece        b         */
	s_check,                    /* check sign         +         */
	s_double_check,             /* double check sign  + (++)    */

	s_count
};

class Notation
	:  public Pointed
{
public:

	int options; 
	/* nt_* constants. */

	String strings[s_count];
	/* no newlines. */

	Notation() {}
	/* Raw */ 


	Notation(char new_chars[10],
		 int new_options,
		 String new_ep,
		 String new_short_castling,
		 String new_long_castling,
		 String check,
		 String double_check);
	/* This constructor is only used to initialize SAN.  Therefore we don't
	 * bother to pass exotic settings. 
	 */

	String Format_Move_Name(Move_Name) const;

	String Format_Piece_Type(int piece_type) const
		{
			assert (piece_type >= 0 && piece_type < 6);
			return strings[s_P + piece_type]; 
		}

	const String &operator [] (int i) const
		{
			assert (i >= 0 && i < s_count);
			return strings[i];
		}
	String &operator [] (int i) 
		{
			assert (i >= 0 && i < s_count);
			return strings[i];
		}
	void Reset();
	static const char *Get_Default_Configuration_Name();
	/* implemented in NOTATION_RESET. */ 

private: 
	void Append_Sign(String &string, 
			 int piece_type) 
		const
		{
			assert (piece_type >= 0 && piece_type < 6); 
			if ((piece_type != pi_P) 
			    ||  (options & (nt_pawn_sign | nt_descriptive)))
				string += Format_Piece_Type(piece_type);
		}
};

extern const int san_chars[26];
/* For each letter from A to Z, the piece type that corresponds to the
 * letter.  If the letter 'x' is a character for a piece in SAN, then
 * san_chars['x'-'a'] contains the piece type.  If the characters is
 * not used in SAN, then the character might correspond to a piece
 * type in another language.  Letters that do not match any languages
 * are indicated by 8. 
 */

Inline int San_Char(char c)
	/* Interface to SAN_CHARS. 
	 */
{
	if (c >= 'A' && c <= 'Z')
		c += ('a' - 'A');

	if (! (c >= 'a' && c <= 'z'))
		return 8;

	return san_chars[c - 'a'];
}

extern Notation san;
/* Standard algebraic notation.  Used in the PGN standard.  Not used for user
 * display. 
 */

#endif /* ! NOTATION_HH */
