#include "process.hh"

#include "error.hh"
#include "format.hh"
#include "filename.hh"
#include "main.hh"
//#include "stream.hh" /* to set the hook */ 

#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>
#include <libgen.h> /* basename(3) on Solaris */ 
#include <fcntl.h>

#ifndef PRIO_MAX
#    ifdef NZERO
#        define PRIO_MAX ((2 * NZERO) - 1)
#    else
#        define PRIO_MAX 19
#    endif
#endif

#if 0
void Process_FD_Set(int fd)
	/* Added as hook to STREAM. */
{
	/* Set close on exec flag on each FD opened.  */
	fcntl(fd, F_SETFD, FD_CLOEXEC);
}

static class Process_Main
{
public:

	Process_Main()
	{
		Stream::Set_FD_Creation_Hook(& Process_FD_Set);
	}
} process_main;
#endif /* 0 */ 

/* If we get a SIGTERM, send of also to our child.  We assume there's only one
 * child at any moment. 
 */
pid_t process_last_pid= -1;

void Process_Term(int)
{
	Verbose_Process("terminating");
	if (process_last_pid < 0)   return; 
	Verbose("\tkill (%u)", process_last_pid);
	kill (process_last_pid, SIGTERM); 
	exit(128+SIGTERM);
}

bool Process_Create(const char *program_filename, 
		    const Vector <String> &args_vector,
		    bool wait,
		    pid_t *process,
		    int input,
		    int output,
		    int error_output,
//		    Pointer <Stream> input,
//		    Pointer <Stream> output,
//		    Pointer <Stream> error_output,
		    bool nice)
{
	assert (wait || process != NULL);
	assert (program_filename);

#ifndef NDEBUG
	Verbose_Process("Process_Create(");
	Verbose("\tprogram_filename = \"%s\",", program_filename); 
	Verbose("\twait = %d,", wait);
	Verbose("\tnice = %d,", nice);
	if (!*args_vector)
		Verbose("\t*args = %d,", *args_vector);
	for (int i= 0;  i < *args_vector; ++i)
		Verbose("\targs[%d] = \"%s\",", i, args_vector[i]()); 
	Verbose("\t)");
#endif /* ! NDEBUG */ 


	/* First check if PROGRAM is executable.  If it is not, only the child
	 * process would know, and this function would return TRUE.
	 * Process_Create() should return false if the program doesn't exist.
	 * Relative filenames cannot be checked.   */

	if (program_filename[0] == '/')
	{
		if (access(program_filename, X_OK))
		{
			Error::Set_Filename(program_filename); 
			return false;
		}
	}
	
	signal(SIGTERM, &Process_Term);

	pid_t pid= fork();
	
	switch (pid)
	{
	case -1: /* error */
		return false;

	case 0: /* child */
	{
		/* Set nice. */
		if (nice)  ::nice(PRIO_MAX);

		/* redirect in/output */
		if (input        >= 0) dup2(input,        0);
		if (output       >= 0) dup2(output,       1);
		if (error_output >= 0) dup2(error_output, 2);

		/* Build argument list  */
		const char *args[*args_vector + 2];
		for (int i= 0;  i < *args_vector;  ++i)
			args[i + 1]= args_vector[i]();
		args[*args_vector + 1]= NULL;

		/* execute */ 
		char *s= strdup(program_filename);
		String base_name= basename(s);
		free(s);
		args[0]= base_name(); 
		execvp(program_filename,
		       (char *const *)args);

		/* execvp() only returns on error */

		fprintf(/*IO::err()*/stderr,
			"\t[child] *** error:  cannot execute \"%s\":  %s.\n",
			program_filename,
			Error::Get()());

		_exit(1);
		/* We must use "_exit()" instead of "exit()", else destructors
		 * would be called in the forked process.   */
		return false;
	}

	default: /* parent */
		Verbose("\t=%u", pid);
		process_last_pid= pid; 
		if (wait)
			waitpid(pid, NULL, 0);
		else
		{
			*process= pid;
		}

		return true;
	}
}
